/* dstart - Start and manage the lifetime of daemons
   Copyright (C) 2012, 2014, 2019 William Haddon

   This program is free software; you can redistribute it and/or modify
   it under the terms of version 2 of the GNU General Public License,
   as published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include <errno.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/wait.h>

#ifndef RUNDIR
#define RUNDIR "/run"
#endif

#define SOCKET_SUBDIR "dstart"

#define EMSG_USAGE "USAGE: %s [-wr] name command\n"
#define EMSG_NOMEM "%s: Out of memory\n"
#define EMSG_FORK "%s: Can not fork\n"
#define EMSG_SOCKET "%s: Can not create socket: %s\n"
#define EMSG_NAME "%s: Service name \"%s\" is not available\n"
#define EMSG_CHECKNAME "%s: Error checking availability of service name: " \
	"%s\nPerhaps you are not root.\n"
#define EMSG_BIND "%s: Can not bind to socket: %s\n"
#define EMSG_LISTEN "%s: Can not listen on socket: %s\n"
#define EMSG_EXEC "%s: Can not start daemon: %s\n"

#define UNIX_PATH_MAX 108

pid_t cpid;
bool always_restart, restart, running, ready;
char *progname;
char *srvname;
char *cmdprog;
char **cmdargs; 
struct sockaddr_un addr;
int listener, fdmax;

/* Check the availability of the service name and set up the socket address
   structure. Return -1 if the name is not available; -2 if another error
   has occurred. */
int check_name(char *name) {
	int s, r, e;
	struct stat st;
	/* Make sure the socket directory exists; create it if it doesn't */
	if (stat(RUNDIR"/"SOCKET_SUBDIR, &st)) {
		if (errno == ENOENT) {
			if (mkdir(RUNDIR"/"SOCKET_SUBDIR, 0644))
				return -2;
		} else {
			return -2;
		}
	} else if (!S_ISDIR(st.st_mode)) {
		return -2;
	}
	/* Try to connect to the socket; if it fails, either because it doesn't
	exist or there's no service at the other end, we're good; delete the
	socket if it exists so that we can call bind() */
	s = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s < 0)
		return -2;
	addr.sun_family = AF_UNIX;
	snprintf(addr.sun_path, UNIX_PATH_MAX, RUNDIR"/"SOCKET_SUBDIR"/%s", name);
	r = connect(s, (struct sockaddr *)&addr, sizeof(struct sockaddr_un));
	e = errno;
	close(s);
	if (r == 0)
		return -1;
	if (e != ECONNREFUSED && e!= ENOTCONN && e != ENOENT)
		return -2;
	unlink(addr.sun_path);
	return 0;
}

/* Create a new child process */
int fork_daemon() {
	int r;
	r = 0;
	errno = 0;
	switch (cpid = fork()) {
		case -1:
			return -1;
		case 0:
			close(listener);
			execvp(cmdprog, cmdargs);
			_exit(1);
		default:
			if (errno)
				r = -1;
			break;
	}
	return r;
}

/* Called when the daemon process dies */
void sigchld_handler() {
	int status;
	/* Reap the zombie */
	wait(&status);
	/* If the daemon process hasn't set up yet, exit with its exit status */
	if (!ready) {
		unlink(addr.sun_path);
		if (WIFEXITED(status))
			exit(WEXITSTATUS(status));
		exit(1);
	}
	/* Restart the daemon process */
	if (restart) {
		restart = always_restart;
		if (fork_daemon() < 0) {
			unlink(addr.sun_path);
			exit(1);
		}
	} else {
		running = false;
	}
}

/* Called when the daemon process signals that it has initialized */
void sigterm_handler() {
	ready = true;
}

/* Called in the terminal-attached process when the daemon signals that it has
	initialized */
void p_sigterm_handler() {
	exit(0);
}

/* Called in the terminal attached process when the daemon exits without
	signalling that it has initialized */
void p_sigchld_handler() {
	int status;
	wait(&status);
	if (WIFEXITED(status))
		exit(WEXITSTATUS(status));
	exit(1);
}

int main(int argc, char **argv) {
	unsigned char action;
	bool _wait;
	int client, i, r, ch;
	fd_set master, copy;
	sigset_t tmpsigset, oldsigset;
	struct sigaction childact, termact;
	/* Process command line */ 
	progname = argv[0];
	_wait = false;
	always_restart = false;
	while ((ch = getopt(argc, argv, "+wr")) != -1) {
		switch (ch) {
			case 'w':
				_wait = true;
				break;
			case 'r':
				always_restart = true;
				break;
			case '?':
				fprintf(stderr, EMSG_USAGE, progname);
				exit(1);
		}
	}
	restart = always_restart;
	if (argc - optind < 2) {
		fprintf(stderr, EMSG_USAGE, progname);
		exit(1);
	}
	srvname = argv[optind];
	cmdprog = argv[optind+1];
	cmdargs = malloc((argc - optind) * sizeof(char *));
	if (cmdargs == NULL) {
		fprintf(stderr, EMSG_NOMEM, progname);
		exit(1);
	}
	memcpy(cmdargs, argv + optind + 1,
		(argc - optind - 1) * sizeof(char *));
	cmdargs[argc-optind-1] = NULL;
	/* Set up socket */
	listener = socket(AF_UNIX, SOCK_STREAM, 0);
	if (listener < 0) {
		fprintf(stderr, EMSG_SOCKET, progname, strerror(errno));
		exit(1);
	}
	r = check_name(srvname);
	if (r == -1) {
		fprintf(stderr, EMSG_NAME, progname, srvname);
		exit(1);
	} else if (r == -2) {
		fprintf(stderr, EMSG_CHECKNAME, progname, strerror(errno));
		exit(1);
	}
	if (bind(listener, (struct sockaddr *)&addr,
			sizeof(struct sockaddr_un)) < 0) {
		fprintf(stderr, EMSG_BIND, progname, strerror(errno));
		exit(1);
	}
	FD_ZERO(&master);
	FD_SET(listener, &master);
	fdmax = listener;
	/* Listen to the socket */
	if (listen(listener, 20) < 0) {
		unlink(addr.sun_path);
		fprintf(stderr, EMSG_LISTEN, progname, strerror(errno));
		exit(1);
	}
	/* If we want to wait for the daemon to set up, we set a handler in the
		terminal attached process so that SIGTERM returns in exiting normally
		(rather than with a "Terminated" status). When the daemon has set
		itself up, this process will receive this signal. Also set a handler
		for SIGCHLD so we can quit if the daemon process quit before it has
		set itself up. */
	if (_wait) {
		termact.sa_handler = p_sigterm_handler;
		sigemptyset(&termact.sa_mask);
		termact.sa_flags = 0;
		sigaction(SIGTERM, &termact, NULL);
		childact.sa_handler = p_sigchld_handler;
		sigemptyset(&childact.sa_mask);
		childact.sa_flags = 0;
		sigaction(SIGCHLD, &childact, NULL);
	}
	/* Separate the monitor process from the terminal attached process */
	switch(fork()) {
		case -1:
			fprintf(stderr, EMSG_FORK, progname);
			exit(1);
		case 0:
			/* The monitor process. */
			setsid();
			break;
		default:
			/* The terminal attached process. If we are waiting for the daemon
				to set up, wait for signals forever. The signal handlers will
				quit when it is time. Otherwise, just quit. */
			if (_wait)
				while (1)
					pause();
			exit(0);
	}
	
	/* This is all in the monitor process. */
	/* Set up signals */
	childact.sa_handler = sigchld_handler;
	sigemptyset(&childact.sa_mask);
	childact.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGCHLD, &childact, NULL);
	sigemptyset(&tmpsigset);
	sigaddset(&tmpsigset, SIGCHLD);
	if (_wait) {
		/* If we are waiting for the daemon to set up, set up a SIGTERM handler
			that forwards the SIGTERM to the terminal attached process so that
			it can quit. */
		termact.sa_handler = sigterm_handler;
		sigemptyset(&termact.sa_mask);
		termact.sa_flags = 0;
		sigaction(SIGTERM, &termact, NULL);
		ready = false;
	}
	/* Create the daemon process */
	running = true;
	if (fork_daemon() < 0) {
		fprintf(stderr, EMSG_EXEC, progname, strerror(errno));
		unlink(addr.sun_path);
		exit(1);
	}
	/* Wait for daemon to set itself up, if desired */
	if (_wait) {
		while (!ready)
			pause();
		/* If ready is true, the daemon has set itself up and we can forward
			SIGTERM to the terminal attached process. */
		kill(getppid(), SIGTERM);
	} else {
		ready = true;
	}
	/* Block SIGCHLD at all times except during the pselect call. This is so
		that SIGCHLD handling doesn't occur in the middle of any of the other
		operations in the main loop */
	sigprocmask(SIG_BLOCK, &tmpsigset, &oldsigset);
	/* The main loop */
	while (running) {
		copy = master;
		if ((r = pselect(fdmax+1, &copy, NULL, NULL, NULL, &oldsigset)) > 0) {
			for (i = 0 ; i <= fdmax; i++) {
				if (FD_ISSET(i, &copy)) {
					if (i == listener) {
						/* Accept the connection */
						client = accept(listener, NULL, NULL);
						if (client >= 0) {
							FD_SET(client, &master);
							if (client > fdmax)
								fdmax = client;
						}
					} else {
						/* Read an action from the socket */
						if (recv(i, &action, 1, 0) > 0) {
							/* The action is a single unsigned byte; the least
							   significant seven bits are the signal that
							   should be sent to the process. If the MSB is
							   set, the process is expected to die, and should
							   not be restarted. If the MSB is cleared, the
							   process should be restarted. */
							/* If the MSB is set, indicate that the process
								should not be restarted */
							if (action & 0x80)
								restart = false;
							else
								restart = true;
							/* Send the signal */
							if (running)
								kill(cpid, action & 0x7f);
						}
						/* Close the socket, whether we received data, the
							connection was shut down, or there was an error */
						close(i);
						FD_CLR(i, &master);
					}
				} else if (r < 0) { /* Return value of pselect */
					if (errno != EINTR) {
						/* Kill the child, stop listening to sockets, delete
							the socket, and exit */
						if (running) {
							kill (cpid, SIGKILL);
							wait(NULL);
						}
						unlink(addr.sun_path);
						exit(1);
					}
				}
			}
		}
	}
	
	/* Clean up - we only get here if the daemon is no longer running and must
		not be restarted */
	/* Delete the socket file */
	unlink(addr.sun_path);
	return 0;
}
