/* getty - Open terminals
   Copyright (C) 2012, 2019 William Haddon

   This program is free software; you can redistribute it and/or modify
   it under the terms of version 2 of the GNU General Public License,
   as published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#define _POSIX_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#define EMSG_NOARGS "USAGE: %s device\n"
#define EMSG_OPEN "%s: Error opening device \"%s\": %s\n"
#define EMSG_SETSID "%s: Error calling setsid(): %s\n"
#define EMSG_SCTTY "%s: Error setting controlling terminal to \"%s\": %s\n"
#define EMSG_REDIRIN "%s: Error redirecting stdin to \"%s\": %s\n"
#define EMSG_REDIROUT "%s: Error redirecting stdout to \"%s\": %s\n"
#define EMSG_REDIRERR "%s: Error redirecting stderr to \"%s\": %s\n"
#define EMSG_READ "%s: Error reading from stdin: %s\n"
#define EMSG_EXEC "%s: Error executing \"%s\"\n: %s\n"

char *procname, *devname;

/* Print an error message and exit */
void error(char *msg, ...) {
	va_list ap;
	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	exit(1);
}

int main(int argc, char **argv) {
	int tty;
	char hostname[HOST_NAME_MAX + 1];
	char username[LOGIN_NAME_MAX + 2];
	char *s;
	/* Process command line arguments */
	procname = argv[0];
	if (argc < 2)
		error(EMSG_NOARGS, procname);
	devname = argv[1];
	/* Start a new session */
	if (setsid() < 0)
		error(EMSG_SETSID, procname, strerror(errno));
	/* Open the tty */
	tty = open(devname, O_RDWR);
	if (tty < 0)
		error(EMSG_OPEN, procname, devname, strerror(errno));
	/* Set tty as the controlling terminal */
	if (ioctl(tty, TIOCSCTTY, 0) < 0)
		error(EMSG_SCTTY, procname, devname, strerror(errno));
	/* Redirect stdin, stdout, and stderr to tty*/
	if (dup2(tty, 0) < 0)
		error(EMSG_REDIRIN, procname, devname, strerror(errno));
	if (dup2(tty, 1) < 0)
		error(EMSG_REDIROUT, procname, devname, strerror(errno));
	if (dup2(tty, 2) < 0)
		error(EMSG_REDIRERR, procname, devname, strerror(errno));
	close(tty);
	/* Read the username */
	if (gethostname(hostname, HOST_NAME_MAX + 1) < 0) {
		strncpy(hostname, "(error detecting hostname)", HOST_NAME_MAX);
		hostname[HOST_NAME_MAX] = 0;
	}
	printf("%s login: ", hostname);
	fflush(stdout);
	if (fgets(username, LOGIN_NAME_MAX + 2, stdin) == NULL)
		error(EMSG_READ, procname, strerror(errno));
	/* Remove the final newline from the username */
	for (s = username; *s && (*s != '\n'); s++);
	if (*s == '\n')
		*s = 0;
	if (*username) {
		/* Execute the login program */
		execlp("login", "login", username, NULL);
		error(EMSG_EXEC, procname, "login", strerror(errno));
	}
}
