/* init - Control the Linux boot process
   Copyright (C) 2012, 2014, 2019 William Haddon

   This program is free software; you can redistribute it and/or modify
   it under the terms of version 2 of the GNU General Public License,
   as published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/reboot.h>
#include <sys/wait.h>

#ifndef ETCDIR
#define ETCDIR "/etc"
#endif

#ifndef BOOTSCRIPT
#define BOOTSCRIPT ETCDIR"/init/boot"
#endif

#ifndef SHUTDOWNSCRIPT
#define SHUTDOWNSCRIPT ETCDIR"/init/shutdown"
#endif

#define EMSG_NOINIT "This program's PID must be 1. It is not intended to" \
" be invoked from a shell.\n"
#define EMSG_FORK "INIT can not fork()\n"
#define EMSG_EXEC "INIT can not execute \"%s\": %s\n"
#define EMSG_WAIT "INIT: no more processes\n"

pid_t shutdownscript = 0;
int rbcmd = 0;

/* Called when some type of shutdown is requested */
void reboot_handler(int signum) {
	switch (signum) {
	/* The choice of signals is fairly arbitrary, except that SIGINT is
		what happens when Ctrl+Alt+Del is pressed, and all of these
		when passed to normal applications cause them to terminate
		by default */
	case SIGUSR1:
		rbcmd = RB_AUTOBOOT; /* reboot */
		break;
	case SIGTERM:
		rbcmd = RB_POWER_OFF; /* power off */
		break;
	case SIGHUP:
		rbcmd = RB_HALT_SYSTEM; /* halt */
		break;
	}
	switch (shutdownscript = fork()) {
	case -1:
		fprintf(stderr, EMSG_FORK);
		break;
	case 0:
		/* Run the shutdown script */
		if (execl(SHUTDOWNSCRIPT, SHUTDOWNSCRIPT, NULL) < 0) {
			fprintf(stderr, EMSG_EXEC, SHUTDOWNSCRIPT,
				strerror(errno));
			/* Wait a while so the user sees the error message
				before we quit and the system shuts down */
			sleep(2);
			exit(1);
		}
		break;
	default:
		break;
	}
}

int main() {
	struct sigaction intact;
	pid_t pid;
	/* Make sure we are PID 1 */
	if (getpid() != 1) {
		fprintf(stderr, EMSG_NOINIT);
		return 1;
	}
	switch (fork()) {
	case -1:
		fprintf(stderr, EMSG_FORK);
		break;
	case 0:
		/* Start the boot process */
		if (execl(BOOTSCRIPT, BOOTSCRIPT, NULL) < 0) {
			fprintf(stderr, EMSG_EXEC, BOOTSCRIPT,
				strerror(errno));
			exit(1);
		}
		break;
	default:
		/* Install signal handlers */
		intact.sa_handler = reboot_handler;
		sigemptyset(&intact.sa_mask);
		intact.sa_flags = 0;
		sigaction(SIGUSR1, &intact, NULL);
		sigaction(SIGTERM, &intact, NULL);
		sigaction(SIGHUP, &intact, NULL);
		/* Change Ctrl+Alt+Del behavior */
		reboot(RB_DISABLE_CAD);
		/* Reap zombies forever */
		while (1) {
			if ((pid = wait(NULL)) < 0 && errno == ECHILD) {
				/* If there are no more children, don't go into a tight loop */
				fprintf(stderr, EMSG_WAIT);
				return 1;
			}
			if (pid == shutdownscript) {
				/* If the process that just died is the shutdown script we
					started, it's time to shut down the computer. */
				/* Protect data */
				sync();
				/* Do it */
				reboot(rbcmd);
			}
		}
	}
}
